@@ -1,5 +1,7 @@
 import os
 import copy
+import logging
+from packaging import version
 from .data_cls import BertDataBunch, InputExample, InputFeatures
 from tqdm.autonotebook import tqdm
 import matplotlib.pyplot as plt
@@ -14,6 +16,7 @@ from .modeling import (
     DistilBertForMultiLabelSequenceClassification,
     CamembertForMultiLabelSequenceClassification,
     AlbertForMultiLabelSequenceClassification,
+    ElectraForMultiLabelSequenceClassification,
 )
 
 from .bert_layers import BertLayerNorm
@@ -53,10 +56,11 @@ from transformers import (
     DistilBertForSequenceClassification,
     DistilBertTokenizer,
     ElectraConfig,
+    ElectraForSequenceClassification,
+    ElectraTokenizer,
 )
 
 from transformers import AutoModelForSequenceClassification, AutoConfig
-from packaging import version
 
 PYTORCH_VERSION = version.parse(torch.__version__)
 
@@ -102,22 +106,28 @@ MODEL_CLASSES = {
         ),
         CamembertTokenizer,
     ),
+    "electra": (
+        ElectraConfig,
+        (ElectraForSequenceClassification, ElectraForMultiLabelSequenceClassification),
+        ElectraTokenizer,
+    ),
 }
-
-try:
-    from apex import amp
-
+if version.parse(torch.__version__) >= version.parse("1.6"):
     IS_AMP_AVAILABLE = True
-except ImportError:
+    from torch.cuda.amp import autocast
+else:
     IS_AMP_AVAILABLE = False
 
-try:
-    from apex.normalization.fused_layer_norm import FusedLayerNorm
-except Exception:
-    from .bert_layers import BertLayerNorm as FusedLayerNorm
 
-
-def load_model(dataBunch, pretrained_path, finetuned_wgts_path, device, multi_label):
+def load_model(
+    dataBunch,
+    pretrained_path,
+    finetuned_wgts_path,
+    device,
+    multi_label,
+    pos_weight,
+    weight,
+):
 
     model_type = dataBunch.model_type
     model_state_dict = None
@@ -135,6 +145,9 @@ def load_model(dataBunch, pretrained_path, finetuned_wgts_path, device, multi_la
     if multi_label is True:
         config_class, model_class, _ = MODEL_CLASSES[model_type]
 
+        model_class[1].pos_weight = pos_weight
+        model_class[1].weight = weight
+
         config = config_class.from_pretrained(
             str(pretrained_path), num_labels=len(dataBunch.labels)
         )
@@ -143,18 +156,9 @@ def load_model(dataBunch, pretrained_path, finetuned_wgts_path, device, multi_la
             str(pretrained_path), config=config, state_dict=model_state_dict
         )
     else:
-        if model_type == "electra":
-            config = ElectraConfig.from_pretrained(
-                str(pretrained_path),
-                model_type=model_type,
-                num_labels=len(dataBunch.labels),
-            )
-        else:
-            config = AutoConfig.from_pretrained(
-                str(pretrained_path),
-                model_type=model_type,
-                num_labels=len(dataBunch.labels),
-            )
+        config = AutoConfig.from_pretrained(
+            str(pretrained_path), num_labels=len(dataBunch.labels)
+        )
         model = AutoModelForSequenceClassification.from_pretrained(
             str(pretrained_path), config=config, state_dict=model_state_dict
         )
@@ -170,7 +174,7 @@ class BertLearner(Learner):
         output_dir,
         metrics,
         device,
-        logger,
+        logger=logging.getLogger(__name__),
         finetuned_wgts_path=None,
         multi_gpu=True,
         is_fp16=True,
@@ -183,13 +187,21 @@ class BertLearner(Learner):
         adam_epsilon=1e-8,
         logging_steps=100,
         freeze_transformer_layers=False,
+        pos_weight=None,
+        weight=None,
     ):
         if is_fp16 and (IS_AMP_AVAILABLE is False):
             logger.debug("Apex not installed. switching off FP16 training")
             is_fp16 = False
 
         model = load_model(
-            dataBunch, pretrained_path, finetuned_wgts_path, device, multi_label
+            dataBunch,
+            pretrained_path,
+            finetuned_wgts_path,
+            device,
+            multi_label,
+            pos_weight,
+            weight,
         )
 
         return BertLearner(
@@ -261,6 +273,8 @@ class BertLearner(Learner):
         self.best_loss = None
         self.state_cacher = None
 
+        self.scaler = torch.cuda.amp.GradScaler() if is_fp16 is True else None
+
         # Freezing transformer model layers
         if freeze_transformer_layers:
             for name, param in self.model.named_parameters():
@@ -301,11 +315,6 @@ class BertLearner(Learner):
         if hasattr(self.model, "module"):
             self.model = self.model.module
 
-        if self.is_fp16:
-            self.model, optimizer = amp.initialize(
-                self.model, optimizer, opt_level=self.fp16_opt_level
-            )
-
         # Get scheduler
         scheduler = self.get_scheduler(
             optimizer, t_total=t_total, schedule_type=schedule_type
@@ -338,45 +347,25 @@ class BertLearner(Learner):
             epoch_step = 0
             epoch_loss = 0.0
             for step, batch in enumerate(progress_bar(train_dataloader, parent=pbar)):
-                self.model.train()
-                batch = tuple(t.to(self.device) for t in batch)
-                inputs = {
-                    "input_ids": batch[0],
-                    "attention_mask": batch[1],
-                    "labels": batch[3],
-                }
-
-                if self.model_type in ["bert", "xlnet"]:
-                    inputs["token_type_ids"] = batch[2]
-
-                outputs = self.model(**inputs)
-                loss = outputs[
-                    0
-                ]  # model outputs are always tuple in pytorch-transformers (see doc)
-
-                if self.n_gpu > 1:
-                    loss = (
-                        loss.mean()
-                    )  # mean() to average on multi-gpu parallel training
-                if self.grad_accumulation_steps > 1:
-                    loss = loss / self.grad_accumulation_steps
-
-                if self.is_fp16:
-                    with amp.scale_loss(loss, optimizer) as scaled_loss:
-                        scaled_loss.backward()
-                    torch.nn.utils.clip_grad_norm_(
-                        amp.master_params(optimizer), self.max_grad_norm
-                    )
-                else:
-                    loss.backward()
-                    torch.nn.utils.clip_grad_norm_(
-                        self.model.parameters(), self.max_grad_norm
-                    )
+                # Run training step and get loss
+                loss = self.training_step(batch)
 
                 tr_loss += loss.item()
                 epoch_loss += loss.item()
                 if (step + 1) % self.grad_accumulation_steps == 0:
-                    optimizer.step()
+                    # gradient clipping
+                    torch.nn.utils.clip_grad_norm_(
+                        self.model.parameters(), self.max_grad_norm
+                    )
+                    if self.is_fp16:
+                        # AMP: gradients need unscaling
+                        self.scaler.unscale_(optimizer)
+
+                    if self.is_fp16:
+                        self.scaler.step(optimizer)
+                        self.scaler.update()
+                    else:
+                        optimizer.step()
                     scheduler.step()
 
                     self.model.zero_grad()
@@ -445,8 +434,42 @@ class BertLearner(Learner):
         else:
             return global_step, tr_loss / global_step
 
+    ### Training Step
+    def training_step(self, batch):
+        self.model.train()
+        batch = tuple(t.to(self.device) for t in batch)
+        inputs = {
+            "input_ids": batch[0],
+            "attention_mask": batch[1],
+            "labels": batch[3],
+        }
+
+        if self.model_type in ["bert", "xlnet"]:
+            inputs["token_type_ids"] = batch[2]
+
+        if self.is_fp16:
+            with autocast():
+                outputs = self.model(**inputs)
+        else:
+            outputs = self.model(**inputs)
+
+        loss = outputs[0]
+
+        if self.n_gpu > 1:
+            loss = loss.mean()
+
+        if self.grad_accumulation_steps > 1:
+            loss = loss / self.grad_accumulation_steps
+
+        if self.is_fp16:
+            self.scaler.scale(loss).backward()
+        else:
+            loss.backward()
+
+        return loss
+
     ### Evaluate the model
-    def validate(self, quiet=False, loss_only=False):
+    def validate(self, quiet=False, loss_only=False, return_preds=False):
         if quiet is False:
             self.logger.info("Running evaluation")
             self.logger.info("  Num examples = %d", len(self.data.val_dl.dataset))
@@ -510,11 +533,15 @@ class BertLearner(Learner):
 
         results = {"loss": eval_loss}
 
+        if return_preds:
+            results["y_preds"] = np.argmax(all_logits.detach().cpu().numpy(), axis=1)
+            results["y_true"] = all_labels.detach().cpu().numpy()
+
         if loss_only is False:
             # Evaluation metrics
             for metric in self.metrics:
                 validation_scores[metric["name"]] = metric["function"](
-                    all_logits, all_labels
+                    all_logits, all_labels, labels=self.data.labels
                 )
             results.update(validation_scores)
 
@@ -523,8 +550,12 @@ class BertLearner(Learner):
     ### Return Predictions ###
     def predict_batch(self, texts=None):
 
+        if self.logger is None:
+            self.logger = logging.getLogger(__name__)
         if texts:
+            self.logger.info("---PROGRESS-STATUS---: Tokenizing input texts...")
             dl = self.data.get_dl_from_texts(texts)
+            self.logger.info("---PROGRESS-STATUS---: Tokenizing input texts...DONE")
         elif self.data.test_dl:
             dl = self.data.test_dl
         else:
@@ -534,6 +565,11 @@ class BertLearner(Learner):
 
         self.model.eval()
         for step, batch in enumerate(dl):
+            self.logger.info(
+                "---PROGRESS-STATUS---: Predicting batch {}/{}".format(
+                    step + 1, len(dl)
+                )
+            )
             batch = tuple(t.to(self.device) for t in batch)
 
             inputs = {"input_ids": batch[0], "attention_mask": batch[1], "labels": None}
@@ -557,10 +593,16 @@ class BertLearner(Learner):
                 all_logits = np.concatenate(
                     (all_logits, logits.detach().cpu().numpy()), axis=0
                 )
+            # self.logger.info(
+            #     "---PROGRESS-STATUS---: Predicting batch {}/{}...DONE".format(
+            #         step + 1, len(dl)
+            #     )
+            # )
 
         result_df = pd.DataFrame(all_logits, columns=self.data.labels)
-        results = result_df.to_dict("record")
+        results = result_df.to_dict(orient="records")
 
+        self.logger.info("---PROGRESS-STATUS---: Predicting batch...DONE")
         return [sorted(x.items(), key=lambda kv: kv[1], reverse=True) for x in results]
 
     # Begin code for LR Finder
@@ -615,11 +657,6 @@ class BertLearner(Learner):
         if hasattr(self.model, "module"):
             self.model = self.model.module
 
-        if self.is_fp16:
-            self.model, self.optimizer = amp.initialize(
-                self.model, self.optimizer, opt_level=self.fp16_opt_level
-            )
-
         self.state_cacher.store("model", self.model.state_dict())
         self.state_cacher.store("optimizer", self.optimizer.state_dict())
 
@@ -694,10 +731,13 @@ class BertLearner(Learner):
             if self.model_type in ["bert", "xlnet"]:
                 inputs["token_type_ids"] = batch[2]
 
-            outputs = self.model(**inputs)
-            loss = outputs[
-                0
-            ]  # model outputs are always tuple in pytorch-transformers (see doc)
+            if self.is_fp16:
+                with autocast():
+                    outputs = self.model(**inputs)
+            else:
+                outputs = self.model(**inputs)
+
+            loss = outputs[0]
 
             if self.n_gpu > 1:
                 loss = loss.mean()  # mean() to average on multi-gpu parallel training
@@ -705,14 +745,7 @@ class BertLearner(Learner):
             loss /= self.grad_accumulation_steps
 
             if self.is_fp16:
-                # For minor performance optimization, see also:
-                # https://nvidia.github.io/apex/advanced.html#gradient-accumulation-across-iterations
-                delay_unscale = ((i + 1) % self.grad_accumulation_steps) != 0
-
-                with amp.scale_loss(
-                    loss, self.optimizer, delay_unscale=delay_unscale
-                ) as scaled_loss:
-                    scaled_loss.backward()
+                self.scaler.scale(loss).backward()
             else:
                 loss.backward()
 
